/* commentmacro.h : C/C++ header file
This header file declares several macros for use in 
debugging builds. The macros are designed to expand to source
code for debugging and to the two slashes "//" text characters,
for release builds. The "//" are interpreted as the beginning of
a single line comment by the C/C++ preprocessor. Note that some 
older C preprocessors do not recognize the "//" character 
sequence as a single line comment, so these macros will not work
for these preprocessors. This design ensures that all lines of 
code using these macros are stripped by the preprocessor and 
completely removed from the release builds. However, the 
greatest advantage of this approach is that any legal expression
can be used to convey debug information, which makes this method
extremely flexible. This method is completely portable.

Use: Assume a debug build where _DEBUG is defined for both the
preprocessor and the compiler. Preceding any line of code with
the DEBUG_ONLY macro will cause that line of code to be compiled
only for debug builds, since in this case the DEBUG_ONLY macro
is defined empty. For release builds the DEBUG_ONLY macro is
interpreted by the preprocessor as "//" so the line is treated
as a single line comment and ignored. For convenience, other
macros are defined which offer variations on this theme, such a
dout (see the code below for their definitions).

This method is intended for use with single lines of code, and
thus, reduces the tedious and unsightly practice of surrounding
a single line with #if/#endif preprocessor directives. It is 
also much more flexible than other parameterized macro 
approaches. For excluding multiple lines of code, the utility of 
preceding each line with the DEBUG_ONLY macro is diminished. In 
this case it would probably be better to resort to the more 
traditional approach:

<code>
#if _DEBUG
// .... several lines of debug code
#endif
<text>

Some "C++" examples:

<code>
DEBUG_ONLY if (x>100) std::cout << "WARNING! x=" << x <<"\n";
dout << "Hello World!\n";
<text>

Some "C" examples:

<code>
DEBUG_ONLY for(x=0;x<10;x++) printf("%c",MyArray[x]);
DEBUG_ONLY printf("Hello World!\n");

__DATE__WHO_##_COMMENT__________________________________________
11Oct00 MET 00 Initial coding by Mark E. Timperley
*/
#ifndef _COMMENTMACRO_H_
#define _COMMENTMACRO_H_

#define COMMENT SPTslash(/)
#define SPTslash(s) /##s

#ifdef _DEBUG // DEBUG BUILD

#ifdef __cplusplus // C++ only
#include <iostream>
#define dout std::cout 
#define doutx std::cout << __FILE__ << __LINE__
#endif // __cplusplus

#define DEBUG_ONLY

#else // RELEASE BUILD

#ifdef __cplusplus // C++ only
#define dout COMMENT
#endif // __cplusplus

#define DEBUG_ONLY COMMENT

#endif // _DEBUG

#endif // _COMMENTMACRO_H_
<text>

