// (C) 2000, Fernando Luis Cacciola Carballal.
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// You are welcome to contact the author at: fcacciola@fibertel.com.ar

#ifndef VARIANT_H
#define VARIANT_H

#include<typeinfo>
#include<string>
#include<stdexcept>

//
// NOTE: The following conditional compilation macros will change the interface
// for variant_t.
//

// NO_EXPLICIT_MEMBER_TPL_INSTANTIATION :
//
// Some compilers fail to explicitely instantiate member template functions:
// They won't compile: int n = my_variant.get<int>()
// One solution is to add an extra dummy argument to help name lookup.
//
// If NO_EXPLICIT_MEMBER_TPL_INSTANTIATION is defined, some methods will have this
// extra argument. This extra argument is not used so it is declared as a const* so
// you can pass (T cosnt*)NULL.
//

#ifdef _MSC_VER
  #define NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
  #pragma warning(disable: 4800) // 'int' : forcing value to bool 'true' or 'false' (performance warning)
#endif

#ifdef __GNUC__
  #define NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
#endif

// Class variant_t
// Encapsulates a value of arbitrary type.
// It can be constructed with values of any type without
// explicit casting. Its value can be retrieved by
// converting it to a copy of apropriate type.
// If an attempt is made to convert it to a type other than
// the type used with the constructor, invalid_argument is
// thrown.
// variant_t objects can be efficiently and safely copied
// without loosing the value and type.
//
// Mar 2000, Fernando Luis Cacciola Carballal.
//
// Changes,
//  26/09/2000 FLC (T v) arguments changed to (T const& v).
//  26/09/2000 FLC std:: added to throw expresion.
//  27/09/2000 FLC MSVC changes (thanks to Jaroslav Bradik).
//                   * NO_EXPLICIT_MEMBER_TPL_INSTANTIATION conditional compilation added.
//                   * One of the 'is_type' methods was renamed to 'is_type_as'
//                       to avoid overload resolution problems.
//  12/10/2000 FLC Modification for compatibility with GNU GCC 2.95.2
//  18/10/2000 FLC type_name() added (thanks to Ivan Erceg),
//                 is_type...() retyped in terms of type_name(), testbed extended. 
//
class variant_t
{
  public :

    variant_t() : data ( NULL ) {}

    // This member template constructor allows you to
    // instance a variant_t object with a value of any type.
    template<typename T> variant_t ( T const& v )
      : data ( new Impl<T>(v) )
      { data->AddRef() ; }

    // NOTE: MSVC requires the template constructor to be declared BEFORE copy contructor.
    variant_t( variant_t const& rhs )
      { if ( rhs.data != NULL )
          rhs.data->AddRef() ;
        data = rhs.data ;
      }
    ~variant_t()
      { if ( data != NULL )
          data->Release() ;
      }

    // NOTE: This code takes care of self-asignment.
    // DO NOT CHANGE THE ORDER of the statements.
    variant_t& operator = ( const variant_t& rhs )
      {
        if ( rhs.data != NULL )
          rhs.data->AddRef();
        if ( data != NULL )
          data->Release();
        data = rhs.data ;
        return * this ;
      }

    // Returns the type name of the type used to construct this instance.
    // Altough this string is implementation-specific, it will be exactly the same
    // as typeid(T).name() where T is the type used to create this variant_t instance.
    std::string type_name() const
      { return data->type_name() ; }

    // This generic conversion operator let you retrieve
    // the value held.
    // To avoid template specialization conflicts, it
    // returns an instance of type T, which will be a COPY
    // of the value contained.
    template<typename T> operator T () const
      { return CastFromBase<T>( data )->data ; }

    // With expressions of the form: int a = my_variant ,
    // the compiler needs to take the template argument T
    // from the type of the left side of the assigment.
    // Most compilers won't do that, so you might need this
    // non-operator form, as in:
    // int a = my_variant.get<int>();
    // This forms returns a REFERENCE and not a COPY, which
    // will be significant in some cases.
    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    template<typename T> const T & get( T const* dmy =NULL ) const
      { return CastFromBase( data , dmy )->data ; }
    #else
    template<typename T> const T & get() const
      { return CastFromBase<T>( data )->data ; }
    #endif

    // This method can be used to test if this variant_t
    // holds a value of type T. It takes no arguments,
    // so it can only be used with a explicit template
    // instancing:
    //   variant_t var(3) ; if ( var.is_type<int>() ) ...
    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    template<typename T> bool is_type( T const* =NULL ) const
    #else
    template<typename T> bool is_type() const
    #endif
      { return type_name()==typeid(T).name() ; }
      
    // This method can be used to test if this variant_t
    // holds a value of type T.
    // It takes one argument, so it can be used with
    // a test variable of the desired type:
    //   int n=3; variant_t var(n) ;
    //   if ( var.is_type_as(n) ) ...
    template<typename T> bool is_type_as(T const& v) const
      { return type_name()==typeid(T).name() ; }

    template<typename T> bool is_type_as(T const* v) const
      { return type_name()==typeid(T).name() ; }

  private :

    // This base class provides polymorphism to Impl<>, due
    // to the virtual destructor.
    // It implements the reference counting over Impl<>.
    struct ImplBase
    {
      ImplBase() : refs ( 0 ) {}
      virtual ~ImplBase() {}
      void AddRef () { refs ++ ; }
      void Release() { refs -- ;
                       if ( refs == 0 )
                         delete this ;
                     }
      virtual std::string type_name() const = 0 ;
      size_t refs ;
    } ;

    // An instance of Impl<T> is just a straight copy of a
    // value of type T, but re-typed as a polymorphic type.
    // As long as the copy constructor for T is properly
    // implemented, the value will be properly copied.
    template<typename T>
    struct Impl : ImplBase
    {
      Impl ( T const& v ) : data ( v ) {}
     ~Impl () {}
      virtual std::string type_name() const { return typeid(T).name() ; } ;
      T data ;
    } ;

    template<typename T>
    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    static Impl<T>* CastFromBase ( ImplBase* v , T const* =NULL )
    #else
    static Impl<T>* CastFromBase ( ImplBase* v )
    #endif
    {
      // This upcast will fail if T is other than the T used
      // with the constructor of variant_t.
      Impl<T>* p = dynamic_cast<Impl<T>*> ( v ) ;
      if ( p == NULL )
        throw std::invalid_argument
         ( typeid(T).name()+std::string(" is not a valid type"));
      return p ;
    }

    ImplBase* data ;
} ;


//
// non-member auxiliary functions.
// Needed for compilers that fail with: my_variant.get<int>().
//

// usage: int n = variant_cast<int>(my_variant);
template<typename T>
inline
T const& variant_cast  (const variant_t& v)
{
  #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    return v.get((T const*)(NULL));
  #else
    // gcc 2.95.2 refuses to parse this
    return v.get<T>();
  #endif
}

// usage: bool b = variant_is_type_as(my_variant,my_type_variable);
//
// NOTE: For some reason MSVC fails to properly instantiate this functions unless
// the 'T const&' argument is added.
template<typename T>
inline
bool variant_is_type_as (const variant_t& v, T const& t)
{
  return v.is_type_as(t);
}
template<typename T>
inline
bool variant_is_type_as (const variant_t& v, T const* t)
{
  return v.is_type_as(t);
}


#endif



