// (C) 2000, Fernando Luis Cacciola Carballal.
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
// You are welcome to contact the author at: fcacciola@fibertel.com.ar

#ifndef VARIANT_CC_H
#define VARIANT_CC_H

#include<typeinfo>
#include<string>
#include<stdexcept>

//
// NOTE: The following conditional compilation macros will change the interface
// for variant_t.
//

// NO_EXPLICIT_MEMBER_TPL_INSTANTIATION :
//
// Some compilers fail to explicitely instantiate member template functions:
// They won't compile: int n = my_variant.get<int>()
// One solution is to add an extra dummy argument to help name lookup.
//
// If NO_EXPLICIT_MEMBER_TPL_INSTANTIATION is defined, some methods will have this
// extra argument. This extra argument is not used so it is declared as a const* so
// you can pass (T cosnt*)NULL.
//

#ifdef _MSC_VER
  #define NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
  #pragma warning(disable: 4800) // 'int' : forcing value to bool 'true' or 'false' (performance warning)
#endif

#ifdef __GNUC__
  #define NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
#endif

// Class variant_cc_t
// This is a variation of variant_t.
// It implements a 'copy on copy' scheme: that is,
// each time a variant_cc_t is copied it internal
// data is copied too.
// This avoids the shared implementation at the cost of
// copying the value held.
//
// Those portions of this code which differ from variant_t
// have a comment.
//
// Mar 2000, Fernando Luis Cacciola Carballal.
//
//  26/09/2000 FLC (T v) arguments changed to (T const& v).
//  26/09/2000 FLC std:: added to throw expresion.
//  27/09/2000 FLC MSVC changes.
//                   * NO_EXPLICIT_MEMBER_TPL_INSTANTIATION conditional compilation added.
//                   * One of the 'is_type' methods was renamed to 'is_type_as'
//                       to avoid overload resolution problems.
//  12/10/2000 FLC Modification for compatibility with GNU GCC 2.95.2
//  18/10/2000 FLC type_name() added (thanks to Ivan Erceg),
//                 is_type...() retyped in terms of type_name(), testbed extended.
//
//
class variant_cc_t
{
  public :

    variant_cc_t() : data ( NULL ) {}

    template<typename T> variant_cc_t ( T const& v )
      : data ( new Impl<T>(v) )
      {}

    // * Copy the internal data instead of sharing it.
    variant_cc_t( const variant_cc_t & rhs )
      : data ( rhs.data != NULL ? rhs.data->clone()
                                : NULL
             )
      {}

    // No need to test for concurrent users.
   ~variant_cc_t()
      { delete data ; }

    // * Copy the internal data instead of sharing it.
    variant_cc_t& operator = ( const variant_cc_t& rhs )
      {
        if ( data != rhs.data )
        {
          delete data ;
          data = ( rhs.data != NULL ? rhs.data->clone()
                                    : NULL
                 ) ;
        }
        return * this ;
      }

    std::string type_name() const
      { return data->type_name() ; }

    template<typename T> operator T () const
      { return CastFromBase<T>( data )->data ; }

    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    template<typename T> const T & get( T const* dmy =NULL ) const
      { return CastFromBase<T>( data , dmy )->data ; }
    #else
    template<typename T> const T & get() const
      { return CastFromBase<T>( data )->data ; }
    #endif

    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    template<typename T> bool is_type( T const* =NULL ) const
    #else
    template<typename T> bool is_type() const
    #endif
      { return type_name()==typeid(T).name() ; }

    template<typename T> bool is_type_as(T const& v) const
      { return type_name()==typeid(T).name() ; }
    template<typename T> bool is_type_as(T const* v) const
      { return type_name()==typeid(T).name() ; }

  private :

    // This version has a clone() method so copies
    // of particular Impl<> can be obtained.
    struct ImplBase
    {
      ImplBase() {}
      virtual ~ImplBase () {}
      virtual  ImplBase* clone() const = 0 ;
      virtual std::string type_name() const = 0 ;
    } ;

    // The clone() is implemented here.
    template<typename T>
    struct Impl : ImplBase
    {
      Impl ( T const& v ) : data ( v ) {}
      virtual ~Impl () {}
      virtual ImplBase* clone() const { return new Impl<T>(data) ; }
      virtual std::string type_name() const { return typeid(T).name() ; } ;
      T data ;
    } ;

    template<typename T>
    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    static Impl<T>* CastFromBase ( ImplBase* v , T const* =NULL )
    #else
    static Impl<T>* CastFromBase ( ImplBase* v )
    #endif
    {
      Impl<T>* p = dynamic_cast<Impl<T>*> ( v ) ;
      if ( p == NULL )
        throw std::invalid_argument
         ( typeid(T).name()+std::string(" is not a valid type"));
      return p ;
    }

    ImplBase* data ;
} ;

//
// non-member auxiliary functions.
// Needed for compilers that fail with: my_variant.get<int>().
//

// usage: int n = variant_cast<int>(my_variant);
template<typename T>
inline
T const& variant_cast  (const variant_cc_t& v)
{
  #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    return v.get((T const*)(NULL));
  #else
    // gcc 2.95.2 refuses to parse this
    return v.get<T>();
  #endif
}

// usage: bool b = variant_is_type_as(my_variant,my_type_variable);
//
// NOTE: For some reason MSVC fails to properly instantiate this functions unless
// the 'T const&' argument is added.
template<typename T>
inline
bool variant_is_type_as (const variant_cc_t& v, T const& t)
{
  return v.is_type_as(t);
}
template<typename T>
inline
bool variant_is_type_as (const variant_cc_t& v, T const* t)
{
  return v.is_type_as(t);
}

#endif
