#include<iostream>
#include<vector>
#include<algorithm>

// gcc2.95 doesn't has <sstream> (yet).
#ifndef __GNUC__
  #include<sstream>
#else
  #include<strstream>
#endif

#ifdef TEST_CC_VERSION
  #include "variant_cc.h"
  typedef variant_cc_t variant_t ;
#else
  #include "variant.h"
#endif

using namespace std;

// test0() shows the basic construction and use of variant_t with
// various types.
void test0()
{
  variant_t _int ( 2 ) ;
  variant_t _dbl ( 3.14 ) ;
  variant_t _str ( string ( "This is a string" ) ) ;
  // IMPORTANT NOTE: The above statement COULD NOT have been
  //  variant_t _str ( "This is a string" ) ;
  // The expression "This is a string" is of type
  //  const char *, which is just a pointer.
  // The copy of a pointer is just another pointer of the
  // same value, not a new membory block with a copy of
  // the contents of the original block.
  // The value copied and stored in _str would be the
  // pointer value and not the character array contents.

  cout << "BEGIN test0" << endl ;
  #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    cout << variant_cast<int>   (_int) << " is of type " << _int.type_name() << endl ;
    cout << variant_cast<double>(_dbl) << " is of type " << _dbl.type_name() << endl ;
    cout << variant_cast<string>(_str) << " is of type " << _str.type_name() << endl ;
  #else
    cout << (int)   _int << " is of type " << _int.type_name() << endl ;
    cout << (double)_dbl << " is of type " << _dbl.type_name() << endl ;
    cout << (string)_str << " is of type " << _str.type_name() << endl ;
  #endif
  cout << "END test0"   << endl << endl ;
}

// test1() show how variant_t can be used as efficient
// return values. The internal data is not copied but shared
// among the various variant_t objects involved.
variant_t test1_aux()
{
  variant_t r ( string ( "This is a string" ) ) ;
  return variant_t(r) ;
}
void test1()
{
  cout << "BEGIN test1" << endl ;
  #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    cout << variant_cast<string>(test1_aux()) << endl ;
  #else
    cout << (string)test1_aux() << endl ;
  #endif
  cout << "END test1" << endl << endl ;
}

// test2() shows an invalid convertion throwing
// invalid_argument.
void test2()
{
  cout << "BEGIN test2" << endl ;
  try
  {
    variant_t _dbl(3.14);
    #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
      char* n = variant_cast<char*>(_dbl) ; (n);
    #else
      char* n = _dbl ; (n);
    #endif
  }
  catch ( invalid_argument& x )
  {
    cout << "exception invalid_argument: " << x.what() << endl ;
  }
  cout << "END test2" << endl << endl ;
}

// test3() shows an array of variant_t
void test3()
{
  cout << "BEGIN test3" << endl ;
  variant_t Array [3];
  Array[0]=2;  // int
  Array[1]=3.14; // double
  Array[2]=string("This is a string");
  #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    cout << variant_cast<int>   (Array[0]) << endl ;
    cout << variant_cast<double>(Array[1]) << endl ;
    cout << variant_cast<string>(Array[2]) << endl ;
  #else
    cout << (int)   Array[0] << endl ;
    cout << (double)Array[1] << endl ;
    cout << (string)Array[2] << endl ;
  #endif
  cout << "END test3" << endl << endl ;
}

// test4() shows a vector<variant_t>
string print ( variant_t const& aVar )
{
  #ifndef __GNUC__
    ostringstream ss ;
  #else
    ostrstream ss ;
  #endif

  #ifdef NO_EXPLICIT_MEMBER_TPL_INSTANTIATION
    if ( variant_is_type_as(aVar,0/*int*/) )
      ss << "int: " << variant_cast<int>(aVar) ;
    else if ( variant_is_type_as(aVar,2.3/*double*/) )
      ss << "double: " << variant_cast<double>(aVar) ;
    else if ( variant_is_type_as(aVar,string()) )
      ss << "string: " << variant_cast<string>(aVar) ;
  #else
    if ( aVar.is_type<int>() )
      ss << "int: " << (int)aVar ;
    else if ( aVar.is_type<double>() )
      ss << "double: " << (double)aVar ;
    else if ( aVar.is_type<string>() )
      ss << "string: " << (string)aVar ;
  #endif

  #ifdef __GNUC__
    ss << ends ;
  #endif

  return ss.str();
}

void test4()
{
  cout << "BEGIN test4" << endl ;
  vector<variant_t> Vector ;
  Vector.push_back ( 2 ) ;
  Vector.push_back ( 3.14 ) ;
  Vector.push_back ( string("This is a string") ) ;
  std::transform ( Vector.begin() ,
                   Vector.end  () ,
                   ostream_iterator<string> ( cout , "\n" ) ,
                   print
                 ) ;
  cout << "END test4" << endl << endl ;
}

void test5()
{
  // The following test shows how cross-variant_t assignment is possible.
  variant_t _char('A');
  variant_t _int(2);
  variant_t _flt(3.14f);
  variant_t _dbl(3.14);

  cout << "BEGIN test5" << endl ;

  cout << _char.type_name() << " _char = " << (int) variant_cast<char>(_char) << endl;
  cout << _int .type_name() << " _int  = " << variant_cast<int>(_int) << endl;
  cout << _flt .type_name() << " _flt  = " << variant_cast<float>(_flt) << endl;
  cout << _dbl .type_name() << " _dbl  = " << variant_cast<double>(_dbl) << endl;

  // These cross assigments propagate the 'inner type' of the variant_t rvalue.
  _int = _char ; // _int has a 'char' now.
  _flt = _int  ; // _flt has a 'char' now.
  _dbl = _flt  ; // _dbl has a 'char' now.

  cout << _char.type_name() << " _char = " << variant_cast<char>(_char) << endl;
  cout << _int .type_name() << " _int  = " << variant_cast<char>(_int)  << endl;
  cout << _flt .type_name() << " _flt  = " << variant_cast<char>(_flt)  << endl;
  cout << _dbl .type_name() << " _dbl  = " << variant_cast<char>(_dbl)  << endl;

  cout << "END test5" << endl << endl ;
}

int main()
{
  #ifdef TEST_CC_VERSION
    cout << "variant_cc_t test bed" << endl ;
  #else
    cout << "variant_t test bed" << endl ;
  #endif

  test0() ;
  test1() ;
  test2() ;
  test3() ;
  test4() ;
  test5() ;

  return 0 ;
}

/*
OUTPUT:

BEGIN test0
2
3.14
This is a string
END test0

BEGIN test1
This is a string
END test1

BEGIN test2
exception invalid_argument: char * is not a valid type
END test2

BEGIN test3
2
3.14
This is a string
END test3

BEGIN test4
int: 2
double: 3.14
string: This is a string
END test4
*/
//=================================================================================
