//
// soaputil.h
//
//	Written by Chris Dix, July 2000
//  Don't drop the SOAP.

#ifndef __SOAP_UTILS_H
#define __SOAP_UTILS_H

#include <string>
#include <map>

using namespace std;

//
// There are four classes defined here as starting points 
// for C++ SOAP development.  There is still a lot of 
// functionality missing, but they show how you can go 
// about composing messages, RPC messages in particular,
// and sending them over HTTP.  
//
// The reason there is missing functionality is because I 
// didn't want to use an XML parser for this example.  Take 
// these as a starting point, but go get a parser.
// Please. :)
//
//      SimpleSoapMessage - 
//          A base class for messages, it builds an 
//      Envelope based on a Body.  It still needs work to 
//      handle Headers and a collection of namespaces, but
//      an XML parser will handle things like that much 
//      better.
//      
//      SimpleSoapRpcMessage -
//          Derives from SimpleSoapMessage, builds the 
//      Body based on a method name and parameter list.
//      Also can parse out a return value from a 
//      response, which it does in the examples.  It cannot
//      handle multiple out params right now, just the 
//      return value, but if you use a parser, this is 
//      trivial.
//      
//      SoapTransport -
//          Send a message over a particular protocol.
//  
//      SoapHttpPostTransport - 
//          Derives from SoapTransport.  It can send a 
//      SimpleSoapMessage over HTTP and get a response.
//      Insert your favorite API here (CSocket, WinInet,
//      whatever you like), I stuck with plain sockets
//      to keep the code more portable.
//

namespace SoapUtils
{
    typedef map<string, string> strmap;
    typedef strmap::const_iterator striterator;

    const string strVersion = 
        "http://schemas.xmlsoap.org/soap/envelope/";
    const string strSoapPrefix = 
        "SOAP-ENV";
    const string strTypeHeader = 
        "Content-Type: text/xml";
    const string strLenHeader = 
        "Content-Length: ";
    const string strHttpContinue = 
        "HTTP/1.1 100";
    const string strHttpOK = 
        "HTTP/1.1 200";
    const string strHttpFault = 
        "HTTP/1.1 500";

    const int nMaxLen = 1024;

    class SimpleSoapMessage
    {
    public:
        SimpleSoapMessage();
        virtual ~SimpleSoapMessage();

        virtual bool putText( const string& strText );
        virtual string getText( void ) const;

        bool containsFault( void ) const
        {
            return( m_bContainsFault );
        }
        virtual bool isValid( void ) const
        {
            return( m_bIsValid );
        }

        virtual void putBody( const string& strBody );
        virtual string getBody( void ) const;

        virtual string getError( void ) const;

        virtual void clear( void );

    protected:
        // XML parsing, extract the element content of the 
        // element named pszName from pszText.  This expects
        // the full name, nsprefix:basename.
        bool extractElementContent( const string& strText, 
            const string& strName, string& strElem );

        // XML parsing, extract the element content of the 
        // first element you find in pszElem.  Both the name 
        // and the content are returned.
        bool extractFirstElement( const string& strElem, 
            string& strName, string& strChild );

        // XML parsing, split a full name into namespace 
        // prefix and basename.
        void splitElementName( const string& strName, 
            string& strPrefix, string& strBasename );

        // Parse the text into the pieces we are 
        // interested in.
        virtual bool parse( const string& strText );

    protected:
        string      m_strPrefix;
        string      m_strBody;
        bool        m_bContainsFault;
        
        bool        m_bIsValid;
        string      m_strError;
    };

    class SimpleSoapRpcMessage : public SimpleSoapMessage
    {
    public:
        SimpleSoapRpcMessage();
        virtual ~SimpleSoapRpcMessage();

        string getBody( void ) const;

        virtual string getMethod( void )
        {
            return( string(m_strMethod) );
        }
        virtual void putMethod( const string& strMethod )
        {
            m_strMethod = strMethod;
        }

        virtual void getMethodNamespace(
            string& strURI,
            string& strPrefix )
        {
            strURI = m_strMethodNamespaceURI;
            strPrefix = m_strMethodNamespacePrefix;
        }
        virtual void putMethodNamespace( 
            const string& strURI,
            const string& strPrefix = "" )
        {
            m_strMethodNamespaceURI = strURI;
            m_strMethodNamespacePrefix = strPrefix;
        }

        void addParameter( const string& strParam,
            const string& strValue );
        bool containsParameter( const string& strParam );
        string getParameterValue( const string& strParam );
        void removeParameter( const string& strParam );

        virtual string getReturnValue( void );

        virtual void clear( void );

    protected:
        virtual bool parse( const string& strText );

    protected:
        string      m_strMethod;
        string      m_strMethodNamespacePrefix;
        string      m_strMethodNamespaceURI;
        
        strmap      m_mapParams;
    };

    class SoapTransport
    {
    public:
        SoapTransport() : m_bDebugging(false)
        {
        }
        virtual ~SoapTransport() 
        {
        }
        
        virtual bool sendMessage( 
            const SimpleSoapMessage* preq,
            SimpleSoapMessage* presp ) = 0;

        bool debugging( void ) const
        {
            return( m_bDebugging );
        }
        void debugging( const bool& bDebugging )
        {
            m_bDebugging = bDebugging;
        }

    protected:
        virtual bool ProtocolInit( void )
        {
            return( true );
        }

        virtual void ProtocolShutdown( void )
        {
        }

    protected:
        bool    m_bDebugging;
    };

    class SoapHttpPostTransport : public SoapTransport
    {
    public:
        SoapHttpPostTransport() : m_uiPort(80)
        {
        }
        virtual ~SoapHttpPostTransport() {}

        virtual bool sendMessage( 
            const SimpleSoapMessage* preq,
            SimpleSoapMessage* presp );

        string getHost( void ) const
        {
            return( m_strHost );
        }
        void putHost( const string& strHost )
        {
            m_strHost = strHost;
        }

        string getAddress( void ) const
        {
            return( m_strAddress );
        }
        void putAddress( const string& strAddress )        
        {
            m_strAddress = strAddress;
        }

        unsigned int getPort( void ) const
        {
            return( m_uiPort );
        }
        void putPort( const unsigned int& uiPort )
        {
            m_uiPort = uiPort;
        }

    protected:
        virtual bool ProtocolInit( void );
        virtual void ProtocolShutdown( void );

    protected:
        string          m_strHost;
        string          m_strAddress;
        unsigned int    m_uiPort;

        string          m_strAction;
    };

};

#endif	// __SOAP_UTILS_H
